from typing import IO, Callable, List, Dict, TextIO
from xml.etree.ElementTree import ElementTree, iterparse
from os import fstat
from tqdm import tqdm

from gpx_lite.gpx import GPX, GPXTrack
from gpx_lite.gpxtrackpoint import GPXTrackPoint
from gpx_lite.gpxtracksegment import GPXTrackSegment
from gpx_lite.utils import parse_xml





class GPXParser:
    """
    Parser for gpx tracks.

    Args:
       file:  file handler

    Usage:

        import gpx-lite as parser
                 ...
        with open(filename, 'r') as gpx_file:
            gpx:GPX = parser.parse(gpx_file)
                  or
            gpx:GPX = parser.iterparse(gpx_file)

    """

    __slots__ = ('_gpx', '_source')

    def __init__(self, file: TextIO)->None:
        self._source: TextIO = file
        self._gpx: GPX = GPX()

    def parse(self, xml_parser: Callable = parse_xml)->GPX:
        """
        Parser for relatively small gpx files.
        By default uses xml.etree.ElementTree.fromstring() method
        inside helper function, that also removes namespace from the root tag.

        :param xml_parser: function returning ElementTree with tags that don't contain namespace
        :return: gpx with loaded data
        """
        xml: ElementTree = xml_parser(self._source.read())
        for trk in xml.iterfind('trk'):
                new_track: GPXTrack = GPXTrack()
                try:
                    new_track.name = trk.find('name').text
                except AttributeError:
                    pass
                try:
                    new_track.number = trk.find('number').text
                except AttributeError:
                    pass
                for seg in trk.iterfind('trkseg'):
                    new_segment: GPXTrackSegment = GPXTrackSegment()
                    for point in seg.iterfind('trkpt'):
                        values: Dict[str, str] = point.attrib
                        try:
                            point.attrib['time'] = point.find('time').text
                        except AttributeError:
                            point.attrib['time'] = None
                        new_point = GPXTrackPoint(float(values['lat']),
                                                  float(values['lon']),
                                                  values['time'])
                        new_segment.append(new_point)
                    new_track.append(new_segment)
                self._gpx.append(new_track)
        return self._gpx

    def iterparse(self)->GPX:
        """
        Incremental reading for large gpx files using xml.etree.ElementTree.iterparse().
        Loads data from filehandler to gpx object.

        :return: gpx with loaded data
        """
        points: List[GPXTrackPoint] = []
        segments: List[GPXTrackSegment]= []
        tracks: List[GPXTrack] = []
        name: str = ''
        number: str = ''
        time: str = ''
        file_size = fstat(self._source.fileno()).st_size
        with tqdm(total=file_size, unit_scale=True, unit='b', desc="Loading gpx") as pbar:
            for _, elem in iterparse(self._source):
                if 'name' in elem.tag:
                    name = elem.text
                elif 'number' in elem.tag:
                    number = elem.text
                elif 'time' in elem.tag:
                    time = elem.text
                elif 'trkpt' in elem.tag:
                    points.append(GPXTrackPoint(float(elem.attrib['lat']),
                                                float(elem.attrib['lon']),
                                                time))
                elif 'trkseg' in elem.tag:
                    segments.append(GPXTrackSegment([p for p in points]))
                    points.clear()
                elif 'trk' in elem.tag:
                    tracks.append(GPXTrack(name, number, [s for s in segments]))
                    segments.clear()
                elem.clear()
                pbar.update(52)
            self._gpx = GPX(tracks=tracks)
        return self._gpx


if __name__ == '__main__':
    from time import process_time
    fn1 = '/home/olga/Documents/GPX/load_test/traces350.gpx' # 360 Mb
    fn2 = '/home/olga/Documents/GPX/traces-raw.gpx'          # 1.3 Gb
    fn3 = '/home/olga/Documents/GPX/liftago.gpx'             # 3.2 Gb
    # with open(fn1, 'r') as xml_file:
    #     parser = GPXParser(xml_file)
    #     gpx = parser.parse()
    # print(gpx)
    # for track in gpx:
    #     print(track)
    #     for seg in track:
    #         print(seg)
    #         for pt in seg:
    #             print(pt)
    #             print(pt.latitude, pt.longitude,pt.time)

    start = process_time()

    with open(fn3, 'r') as fh:
        parser = GPXParser(fh)
        gpx = parser.iterparse()
    np = sum(map(lambda tr: len(tr[0]), gpx))
    print('Load gpx, %s points at %.2f sec' % (np, process_time() - start))
    #              ,
    # with open('/home/olga/Documents/GPX/save_raw.gpx', 'w') as fh2:
    #     print('Handler ', type(fh2))
    #     gpx.write_to_file(fh2)
    # print(process_time()-start)
    # print(gpx)



from typing import Union, Optional, List, Iterator, Iterable, IO
from copy import deepcopy

from gpx_lite.gpxtrackpoint import GPXTrackPoint
from gpx_lite.gpxtracksegment import GPXTrackSegment


class GPXTrack:
    """
    Attributes:
        name: track name, str or None
        number: track number, int or None
        segments: list of GPXTrackSegments
    """
    __slots__ = ('_name', '_number', '_segments')

    def __init__(self, name: Optional[str]=None,
                 number: Optional[Union[str, int]]=None,
                 segments: Optional[List[GPXTrackSegment]]=None):
        self._name: Optional[str] = name
        if isinstance(number, int):
            self._number: Optional[int] = number
        else:
            try:
                self._number: Optional[int] = int(number)
            except TypeError:
                self._number: Optional[int] = None
        self._segments: List[GPXTrackSegment] = segments if segments else []

    def __repr__(self)->str:
        return '<GPXTrack %s %s [.%s segments.]>' \
               % (self._name, self._number, len(self._segments))

    def __len__(self)->int:
        return len(self._segments)

    def __iter__(self)->Iterator[GPXTrackSegment]:
        return iter(self._segments)

    def __getitem__(self, key: int)->Union[GPXTrackSegment,
                                           List[GPXTrackSegment]]:
        if isinstance(key, int):
            return self._segments[key]
        elif isinstance(key,slice):
            return self._segments[key.start:key.stop:key.step]
        else:
            raise TypeError('Index must be int, not {}'.
                            format(type(key).__name__))

    def __contains__(self, item: GPXTrackSegment)->bool:
        return item in self._segments

    @property
    def name(self)->str:
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @property
    def number(self)->int:
        return self._number

    @number.setter
    def number(self, num: str):
        self._number = int(num)

    @property
    def segments(self)->List[GPXTrackSegment]:
        return self._segments

    @segments.setter
    def segments(self, segments: List[GPXTrackSegment])->None:
        self._segments = segments

    @property
    def points(self)->List[GPXTrackPoint]:
        """

        :return: list of all points from all the segments
        """
        return [pt for seg in self._segments for pt in seg.points]

    def get_points_no(self)->int:
        """

        :return: total number of points in all segments
        """
        return sum(map(lambda seg: len(seg), self._segments))

    def append(self, item: GPXTrackSegment):
        self._segments.append(item)

    def extend(self, items: Iterable[GPXTrackSegment]):
        self._segments.extend(items)

    def remove(self,item: GPXTrackSegment):
        self._segments.remove(item)

    def remove_empty(self)->None:
        """
        Removes empty segmets from the track
        :return:
        """
        self._segments = [seg for seg in filter(
            lambda seg: len(seg) > 0, self._segments)]

    def _write_to_file(self, fh: IO)->None:
        result: List[str] = ['\n<trk>', ]
        if self._name:
            result.extend(['\n<name>', self._name, '</name>'])
        if self._number is not None:
            result.extend(['\n<number>', str(self._number), '</number>'])
        for string in result:
            fh.write(string)
        for seg in self._segments:
            seg._write_to_file(fh)
        fh.write('\n</trk>')

    def clone(self)->'GPXTrack':
        return deepcopy(self)


if __name__ == '__main__':

    from gpx_lite.gpxtrackpoint import GPXTrackPoint as TrackPoint

    x = 50.0164596
    y = 14.4547907
    p1 = TrackPoint(x, y, '2017-11-22T11:03:36Z')
    p2 = TrackPoint(y, x,'2017-11-22T08:03:36Z')
    p3 = TrackPoint(y,y, '2017-11-13T05:11:09Z')
    p4 = TrackPoint(x, x, '2017-11-22T09:03:36Z')
    seg1 = GPXTrackSegment([p1, p2, p3])
    seg2 = GPXTrackSegment([p2, p3, p4])
    track = GPXTrack('800003627_337', '0')
    print('Empty track with name and number: ', track)
    track.append(seg1)
    print('1 segment added, len = ',  len(track))
    seg3 = GPXTrackSegment([p4, p1])
    track.extend([seg2, seg3])
    print('2 more segments added, len: ', len(track))
    print('Points in all segments: ', track.get_points_no())
    print('Slice: ', track[2:])
    print('Iterator')
    for s in track:
        print(s)

    track.remove(seg1)
    print('1 segment removed: ', len(track))



from datetime import datetime
from typing import Callable
from re import sub
import xml.etree.ElementTree as ET


def parse_time(string: str, parser: Callable = datetime.strptime)->datetime:
    """
    :param string: date and time as a string
    :param parser: function to convert string to datetime
    :return: datetime.datetime
    """
    date_formats = ["%Y-%m-%dT%H:%M:%S.%fZ",
                    '%Y-%m-%dT%H:%M:%SZ']
    for df in date_formats:
        try:
            return parser(string, df)
        except ValueError:
            pass
    raise ValueError('Invalid time format in string %s' % string)


def parse_xml(xml_string: str, parser: Callable=ET.fromstring)->ET.ElementTree:
    """
    Helper function to remove namespace and read ElementTree from string.
    
    :param xml_string: xml represented as a single string
    :param parser: function to read xml from string
    :return: ElementTree
    """
    xml_string = sub(r'\sxmlns="[^"]+"', '', xml_string, count=1) #otherwise ns will appear in tags
    return parser(xml_string)


if __name__ == '__main__':
    fn = '/home/olga/Documents/GPX/load_test/traces10.gpx'
    with open(fn, 'r') as xml_file:
        root = parse_xml(xml_file.read())
        print(root)






from typing import IO
from .gpx import GPX

name = ' gpx_lite '
__all__ = ['gpx.py', 'gpxtrack.py', 'gpxtracksegment.py', 'gpxtrackpoint.py']


def parse(file: IO)->GPX:

    """
    Wrapper fo GPXParser.parse(),
    loads gpx from xml file.

    :param file: file handler
    :return: gpx loaded from xml
    """
    from . import parser

    parser = parser.GPXParser(file)
    return parser.parse()


def iterparse(file: IO)->GPX:

    """
    Wrapper fo GPXParser.iterparse(),
    loads gpx from xml file.

    :param file: file handler
    :return: gpx loaded from xml
    """

    from . import parser

    parser = parser.GPXParser(file)
    return parser.iterparse()



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="gpx-lite",
    version="0.0.5",
    author='Olga Kholkovskaia',
    author_email='olga.kholkovskaia@gmail.com',
    license='Apache License, Version 2.0',
    description="Simple parser for gpx tracks, segments, and points with latitude, longitude, and time.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/aicenter/gpx_lite',
    packages=['gpx_lite'],
    install_requires=['typing>=3.6.2','tqdm'],

)
